#include "queue.h"                            // ns-2
#include "delay.h"                            // ns-2
#include "agent.h"                            // ns-2

#include <list>                               // list
#include <map>                                // multimap

using namespace std;

class InstVar {
public:
	static double bw_atof(const char* s);
	static double time_atof(const char* s);
};

namespace VIFQ {

typedef long long PacketTime;

struct PacketSlot {
	PacketTime serviceTime;
	int size;
	PacketSlot(const PacketTime& t,  int s  ) : serviceTime(t), size(s) {}
};
typedef list<PacketSlot> SlotQueue;

struct Flow {
	PacketTime targetDelay;
	SlotQueue slotQueue;
	int credit;
};

struct FlowSlot {
	PacketTime serviceTime;
	Flow& flow;
	FlowSlot(const PacketTime& t,  Flow& f ) : serviceTime(t), flow(f) {}
};
typedef list<FlowSlot> FlowQueue;

typedef multimap<PacketTime,Packet*> ServiceQueue;

class VIFQueue : public Queue {
	// configuration
	static const int flowCount = 64;
	Flow flowArray[flowCount];
	int linkRate;
	int queueLimit;

	// scheduling
	ServiceQueue packetQueue;
	FlowQueue flowQueue;
	PacketTime currentTime;
	PacketTime serviceTime;

	// tracing
	Tcl_Channel tchan_;

public:
	VIFQueue() : linkRate(0), queueLimit(0), tchan_(0) {}

	virtual void reset() {
		                                             assert( packetQueue.empty() );
		flowQueue.clear();
		for (int i = 0; i < flowCount; i += 1) {
			flowArray[i].slotQueue.clear();
			flowArray[i].credit = 0;
		}
		currentTime = 0;
		serviceTime = 0;
	}

	virtual void enque( Packet* p ) {
		Flow& f = flowArray[HDR_IP(p)->prio() % flowCount];
		int plen = HDR_CMN(p)->size();

		// slot admission test
		if ( queueLimit == 0 || serviceTime + plen <= currentTime + queueLimit ) {
			f.slotQueue.push_back( PacketSlot( serviceTime, plen ) );
			f.credit += plen;
			flowQueue.push_back( FlowSlot( serviceTime, f ) );
			serviceTime += plen;
			trace( 'V', p );
		}

		// clean up old slots
		while ( flowQueue.front().serviceTime < serviceTime - queueLimit ) {
                                                  assert( !flowQueue.empty() );
			Flow& g = flowQueue.front().flow;
			if ( !g.slotQueue.empty() && flowQueue.front().serviceTime == g.slotQueue.front().serviceTime ) {
				g.credit -= g.slotQueue.front().size;
				currentTime += g.slotQueue.front().size;
				g.slotQueue.pop_front();
			}
			assert( g.slotQueue.empty() || flowQueue.front().serviceTime < g.slotQueue.front().serviceTime );
			flowQueue.pop_front();
		}

		// packet admission test, reconcile different packet lengths
		if ( f.credit >= plen && ( f.targetDelay == 0 || f.slotQueue.front().serviceTime <= currentTime + f.targetDelay )) {
			f.credit -= plen;
			packetQueue.insert( pair<PacketTime,Packet*>(f.slotQueue.front().serviceTime, p) );
			while ( plen > 0 ) {
			                                          assert( !f.slotQueue.empty() );
				if ( plen < f.slotQueue.front().size ) {
					f.slotQueue.front().size -= plen;
					plen = 0;
				} else {
					plen -= f.slotQueue.front().size;
					f.slotQueue.pop_front();
				}
			}
			// accept packet and trace output
			HDR_CMN(p)->timestamp() = Scheduler::instance().clock();
			trace( 'A', p );
		} else {
			trace( 'D', p );
			drop(p);
		}

	}

	virtual Packet* deque() {
		Packet* p = NULL;
		if ( !packetQueue.empty() ) {
			p = (*packetQueue.begin()).second;
			packetQueue.erase( packetQueue.begin() );
			trace( 'S', p );
			currentTime += HDR_CMN(p)->size();
		} else {
			reset();
		}
		return p;
	}
	virtual int command( int argc, const char*const* argv );
	virtual void trace( char, Packet* );
};

int VIFQueue::command( int argc, const char*const* argv ) {
	Tcl& tcl = Tcl::instance();
	switch (argc) {
	case 3:
		if (strcmp(argv[1], "attach") == 0) {
			int mode;
			tchan_ = Tcl_GetChannel( tcl.interp(), (char*)argv[2], &mode );
			if ( !tchan_ ) {
				tcl.resultf( "VIFQ: can't attach trace %s for writing", argv[2] );
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
		if (strcmp(argv[1], "target") == 0) {
			LinkDelay* del = (LinkDelay*)TclObject::lookup( argv[2] );
			if ( del == 0 ) {
				tcl.resultf( "VIFQ: no LinkDelay object %s", argv[2] );
				return (TCL_ERROR);
			}
			// second call "target" comes from flow monitor attachment with 0 BW
			if ( !linkRate ) linkRate = int( del->bandwidth() / 8 );
			return Queue::command( argc, argv );
		}
		if (strcmp(argv[1], "set-queue-limit") == 0) {
		  if ( !linkRate ) {
		    tcl.resultf( "VIFQ: link rate not set" );
		    return (TCL_ERROR);
      }
			queueLimit = linkRate * InstVar::time_atof(argv[2]);
//      fprintf( stderr, "VIFQ: set queue limit %d", queueLimit );
			return (TCL_OK);
		}
	case 4:
		if (strcmp(argv[1], "add-class") == 0) {
			int cl = atoi(argv[2]);
			flowArray[cl].targetDelay = PacketTime(linkRate * InstVar::time_atof(argv[3]));
//			fprintf( stderr, "VIFQ: add class %d with delay %lld bytes\n", cl, flowArray[cl].targetDelay );
			return (TCL_OK);
		}
	}
	return Queue::command(argc, argv);
}

void VIFQueue::trace( char op, Packet* p ) {
	if (!tchan_) return;
	char wrk[512];
	Flow& f = flowArray[HDR_IP(p)->prio() % flowCount];
	sprintf( wrk, "V %c %4u %10.6f %14llu %6i %3zu %3zu %10.6f\n",
		op, HDR_IP(p)->prio(), Scheduler::instance().clock(), currentTime,
		int(serviceTime - currentTime), packetQueue.size(), f.slotQueue.size(),
		Scheduler::instance().clock() - HDR_CMN(p)->timestamp()
	);
	Tcl_Write( tchan_, wrk, strlen(wrk) );
}

// hook into Tcl for object creation
static class VIFQClass : public TclClass {
public:
	VIFQClass() : TclClass("Queue/VIFQ") {}
	TclObject* create(int, const char*const*) { return new VIFQueue; }
} class_vifq;

};
