#include "ns3/core-module.h"
#include "ns3/point-to-point-module.h"        // PointToPointHelper
#include "ns3/point-to-point-layout-module.h" // PointToPointDumbbellHelper
#include "ns3/internet-module.h"              // Ipv4GlobalRoutingHelper
#include "ns3/applications-module.h"          // ApplicationContainer
#include "ns3/network-module.h"               // InetSocketAddress
#include "ns3/flow-monitor-module.h"
#include "ns3/bulk-send-application.h"
#include "ns3/onoff-application.h"
#include "ns3/log.h"
#include "ns3/dsf-queue.h"
#include <sstream>
#include <fstream>

NS_LOG_COMPONENT_DEFINE("DSF");

#define DSFDCE 1

#if DSFDCE
#include "ns3/dce-module.h"
#include "ns3/linux-stack-helper.h"
#endif

using namespace ns3;
using namespace std;

static const uint64_t mega = 1000000;

static PointToPointDumbbellHelper* topology = NULL;
static ofstream completionFile;
static list<DSFClass> classList;

static double   baseRate = 100;
static uint64_t linkDelay = 1;
static uint64_t	queueTime = 0;
static double   scale = 1.0;
static uint64_t stopTime = 60;
static bool     linuxStack = false;

static int finiteCount = 0;

struct DSF_Desc {
  unsigned int descnum;
  unsigned int classnum;
  double load;
  unsigned int count;
  double start;
  double wait;
  double duration;
  enum Type { File, File1, CBR, Poisson, Pareto, CTrace } type;
  unsigned int amount; // used as packet size for UDP
  double shape;
  unsigned int fileCount;
  unsigned int rtt;
  ifstream* trace;
  uint64_t rateSum;
  uint64_t rateSum2;
  DSF_Desc() : shape(0), fileCount(0), rtt(0), trace(NULL) {}
};

static list<DSF_Desc> descList;

class DSF_App {
  DSF_Desc& tdesc;
  int nodeNum;
  int portNum;
  int renewal;
  Ptr<ParetoRandomVariable> paretoRNG;
  Ptr<UniformRandomVariable> uniformRNG;
  Ptr<Application> source;
  Ptr<PacketSink> sink;
  uint64_t fileSize;
  uint64_t filesStarted;
  uint64_t bytesTransmitted;
  double startTime;
public:
  DSF_App( DSF_Desc& tdesc, int n ) : tdesc(tdesc), nodeNum(n), portNum(1024),
    renewal(0), filesStarted(0), bytesTransmitted(0) {
    paretoRNG = CreateObject<ParetoRandomVariable>();
    uniformRNG = CreateObject<UniformRandomVariable>();
    // DCE Linux must initialize first at time 0.0
    if (tdesc.start == 0.0) tdesc.start = 0.000001;
    filesStarted += 1;
    Simulator::Schedule(Seconds(tdesc.start), DSF_App::Start, this);
  }
  static void Start(void *ptr) {
    DSF_App& This = *(DSF_App*)ptr;
    string proto;
    switch (This.tdesc.type) {

    case DSF_Desc::File:
    case DSF_Desc::File1: {
#if DSFDCE
      if (linuxStack) proto = "ns3::LinuxTcpSocketFactory";
      else
#endif
        proto = "ns3::TcpSocketFactory";
      if (This.tdesc.rtt > 0) {
        if (This.tdesc.rtt <= linkDelay * 4) {
          cerr << "ERROR: RTT " << This.tdesc.rtt << " impossible with link delay " << linkDelay << endl;
          exit(1);
        }
        topology->GetLeft(This.nodeNum)->GetDevice(0)->GetObject<PointToPointNetDevice>()->GetChannel()->SetAttribute("Delay", TimeValue(MilliSeconds((This.tdesc.rtt - linkDelay/4) / 2)));
      }
      Ptr<BulkSendApplication> app = CreateObject<BulkSendApplication>();
      app->SetAttribute( "Protocol", StringValue(proto) );
      if (This.tdesc.shape) {
        This.fileSize =  This.paretoRNG->GetValue(This.tdesc.amount, This.tdesc.shape, INFINITY);
      } else {
        This.fileSize = This.tdesc.amount;
      }
      app->SetAttribute( "MaxBytes", UintegerValue(This.fileSize) );
      app->SetTOS(This.tdesc.classnum << 2);
      app->setDoneCallback(Done, ptr);
      app->setStopCallback(Stop, ptr);
      This.source = app;
    } break;

    case DSF_Desc::CTrace: {
#if DSFDCE
      if (linuxStack) proto = "ns3::LinuxTcpSocketFactory";
      else
#endif
        proto = "ns3::TcpSocketFactory";
      unsigned int classnum;
      unsigned int rtt;
      if (!(*This.tdesc.trace >> This.fileSize >> rtt >> classnum)) return;
      if (rtt <= linkDelay * 4) {
        cerr << "ERROR: RTT " << rtt << " impossible with link delay " << linkDelay << endl;
        exit(1);
      }
      topology->GetLeft(This.nodeNum)->GetDevice(0)->GetObject<PointToPointNetDevice>()->GetChannel()->SetAttribute("Delay", TimeValue(MilliSeconds((rtt - linkDelay/4) / 2)));
      Ptr<BulkSendApplication> app = CreateObject<BulkSendApplication>();
      app->SetAttribute( "Protocol", StringValue(proto) );
      app->SetAttribute( "MaxBytes", UintegerValue(This.fileSize) );
      app->SetTOS(This.tdesc.classnum << 2);
      app->setDoneCallback(Done, ptr);
      app->setStopCallback(Stop, ptr);
      This.source = app;
    } break;

    // (*) Superposition of n Pareto OnOff Processes
    //  according to [Self-similarity through high-variability: statistical analysis of Ethernet LAN traffic at the source level, ToN 1997]
    //  picking both the same distributions is fine
    //  and in that case the load of each OnOff is simply CBR/2
    default:
      unsigned int pktsize = This.tdesc.amount;
      if (pktsize <= 30) pktsize = This.uniformRNG->GetValue(64, 1500);
      double load = This.tdesc.load / This.tdesc.count;
//      double ival = double(pktsize) / ((baseRate/8) * load); // packet time
      double ival = 0.1;
      ostringstream on, off;
      switch (This.tdesc.type) {
      case DSF_Desc::CBR:
        on << "ns3::ConstantRandomVariable[Constant=1.0]";
        off << "ns3::ConstantRandomVariable[Constant=0.0]";
        break;
      case DSF_Desc::Poisson:
        load = load * 2; // on/off
        on << "ns3::ConstantRandomVariable[Constant=" << ival << ']';
        off << "ns3::ExponentialRandomVariable[Mean=" << ival << ']';
        break;
      case DSF_Desc::Pareto:
        load = load * 2; // on/off
        on << "ns3::ParetoRandomVariable[Shape=" << This.tdesc.shape << "|Mean=" << ival << ']';
        off << "ns3::ParetoRandomVariable[Shape=" << This.tdesc.shape << "|Mean=" << ival << ']';
        break;
      default: terminate( "unknown traffic type", This.tdesc.type );
      }
      int udpSize = pktsize - 30; // UDP payload size (excl. UDP/PPP header)
      double udpRate = udpSize * (baseRate / pktsize);
#if DSFDCE
      if (linuxStack) proto = "ns3::LinuxUdpSocketFactory";
      else
#endif
        proto = "ns3::UdpSocketFactory";
      Ptr<OnOffApplication> app = CreateObject<OnOffApplication>();
      app->SetAttribute( "Protocol", StringValue(proto) );
      app->SetAttribute( "OnTime", StringValue(on.str()) );
      app->SetAttribute( "OffTime", StringValue(off.str()) );
      app->SetAttribute( "PacketSize", UintegerValue(udpSize) );
      app->SetAttribute( "DataRate", DataRateValue(udpRate * load) );
      app->SetTOS(This.tdesc.classnum << 2);
      This.source = app;
    }

    InetSocketAddress sinkAddress(topology->GetRightIpv4Address(This.nodeNum), This.portNum);
    This.renewal += 1;
    This.portNum += 1;
    if (This.portNum >= 65536) This.portNum = 1024;
    This.sink = CreateObject<PacketSink>();
    This.sink->SetAttribute( "Protocol", StringValue(proto) );
    This.sink->SetAttribute( "Local", AddressValue(sinkAddress) );
    This.sink->SetAttribute( "StartTime", TimeValue(Seconds(0)) );
    This.sink->SetNode(topology->GetRight(This.nodeNum));
    Simulator::ScheduleWithContext(topology->GetRight(This.nodeNum)->GetId(),
      Seconds(0), &Application::Initialize, This.sink);

    double start = This.uniformRNG->GetValue(0, This.tdesc.wait);
    This.source->SetAttribute( "StartTime", TimeValue(Seconds(start)) );
    This.source->SetAttribute( "Remote", AddressValue(sinkAddress) );
    if (This.tdesc.duration != 0.0) {
      This.source->SetAttribute( "StopTime", TimeValue(Seconds(start + This.tdesc.duration)) );
    } else {
      This.source->SetAttribute( "StopTime", TimeValue(Seconds(double(stopTime) - Now().GetSeconds())) );
    }
    This.source->SetNode(topology->GetLeft(This.nodeNum));
    Simulator::ScheduleWithContext(topology->GetLeft(This.nodeNum)->GetId(),
      Seconds(0), &Application::Initialize, This.source);
    This.startTime = Now().GetSeconds() + start;
  }

  void Summary() {
    if (completionFile.good()) {
      char buf[512];
      sprintf(buf, "T     SUMMARY TC %-3i N %-3i %lu %11.6f",
        tdesc.descnum, nodeNum, bytesTransmitted,
        bytesTransmitted / Simulator::Now().GetSeconds() );
      completionFile << buf << std::endl;
    }
  }

  static void Done(void *ptr) {
    DSF_App& This = *(DSF_App*)ptr;
    This.sink->ForceStopApplication();
    if (completionFile.good()) {
      char buf[512];
      sprintf(buf, "t %11.6f tc %-3i n %-3i r %-4i p %-4i %11.6f %lu",
        Simulator::Now().GetSeconds(), This.tdesc.descnum,
        This.nodeNum, This.renewal, This.portNum,
        Now().GetSeconds() - This.startTime, This.fileSize);
      completionFile << buf << std::endl;
    }
    This.bytesTransmitted += This.fileSize;
    if (This.tdesc.type == DSF_Desc::File) {
      if (This.tdesc.fileCount != 0) {
        if (This.filesStarted < This.tdesc.fileCount) {
          This.filesStarted += 1;
        } else {
          This.Summary();
          finiteCount -= 1;
          if (finiteCount == 0) Simulator::Stop(Seconds(0));
          return;
        }
      }
      Simulator::Schedule( Seconds(0), &DSF_App::Start, ptr );
    } else if (This.tdesc.type == DSF_Desc::CTrace) {
      Simulator::Schedule( Seconds(0), &DSF_App::Start, ptr );
    }
  }

  static void Stop(void *ptr, uint32_t totBytes) {
    DSF_App& This = *(DSF_App*)ptr;
    if (completionFile.good()) {
      char buf[512];
      sprintf(buf, "t %11.6f tc %-3i n %-3i r %-4i p %-4i %11.6f %u",
        Simulator::Now().GetSeconds(), This.tdesc.descnum,
        This.nodeNum, This.renewal, This.portNum,
        Now().GetSeconds() - This.startTime, totBytes);
      completionFile << buf << std::endl;
    }
    This.bytesTransmitted += totBytes;
    This.Summary();
  }

};

static unsigned int parseConfigFile( const string& filename ) {
  unsigned int ncount = 0;
  unsigned int dcount = 0;
  ifstream ifs(filename.c_str());
  if (!ifs.good()) terminate( "Cannot access config file:", filename );
  string line;
  while (getline(ifs, line)) {
    istringstream iss(line);
    string key;
    iss >> key;
    if ( key[0] == '#' || key.empty() ) continue;
    else if ( key == "END" ) break;
    else if ( key == "class" ) {
      DSFClass c;
      iss >> c.num >> c.delay;
      classList.push_back(c);
    } else if ( key == "flow" ) {
      DSF_Desc t;
      t.descnum = dcount;
      dcount += 1;
      string type;
      iss >> t.classnum >> t.load >> t.count >> t.start >> t.wait >> t.duration
          >> type >> t.amount;
      t.count = t.count * scale;
           if ( type == "cbr" )      t.type = DSF_Desc::CBR;
      else if ( type == "poisson" )  t.type = DSF_Desc::Poisson;
      else if ( type == "pareto" ) { t.type = DSF_Desc::Pareto; iss >> t.shape; }
      else if ( type == "file1" )  { t.type = DSF_Desc::File1;  iss >> t.shape >> t.rtt; }
      else if ( type == "file" )   { t.type = DSF_Desc::File;   iss >> t.shape >> t.fileCount >> t.rtt; }
      else if ( type == "ctrace" ) {
        t.type = DSF_Desc::CTrace;
        string filename;
        iss >> filename;
        t.trace = new ifstream(filename.c_str());
        if (!t.trace->good()) terminate( "problem with file", filename );
      }
      else terminate( "unknown flow type", type );
      if (t.fileCount) finiteCount += t.count;
      ncount += t.count;
      descList.push_back(t);
    } else {
      terminate( "unknown keyword", key );
    }
    if ( iss.fail() || !iss.eof() ) terminate( "error in line (whitespace at end?): ", line );
  }
  return ncount;
}

static void printTime(void) {
  cerr << "simulation time: " << 100 * Simulator::Now() / Seconds(stopTime) << '%' << (char)0x0D;
  Simulator::Schedule(Seconds(stopTime) / 100, &printTime);
}

#if DSFDCE
static void Sysctl(std::string key, std::string value) {
  std::cout << key << "=" << value;
}
#endif

int main (int argc, char *argv[]) {
  NS_LOG_UNCOND ("DSF Simulator");

  // default parameters & command-line overrides
  string    compFile;
  string    configFile  = "dsf.cfg";
  string    flowmonFile;
  string    monFile;
  double    monInterval = 1.0;
  double    monStart    = 1.0;
  int       quiet       = 0;
  uint64_t  run         = 0;
  uint32_t  seed        = 0;
  string    traceFile;
  int       setLD       = 0;
  int       setABE      = 0;
  int       setPRIO     = 0;

  CommandLine cmd;
  cmd.AddValue("baseRate",    "bottleneck rate in Mbps (100)", baseRate);
  cmd.AddValue("compFile",    "completion trace file name",    compFile);
  cmd.AddValue("configFile",  "config file name (dsf.cfg)",    configFile);
  cmd.AddValue("linuxStack",  "use linux stack (0)",           linuxStack);
  cmd.AddValue("flowmonFile", "flowmon file name",             flowmonFile);
  cmd.AddValue("linkDelay",   "link delay in ms (1)",          linkDelay);
  cmd.AddValue("monFile",     "monitor file name",             monFile);
  cmd.AddValue("monStart",    "monitor start in s (1.0)",      monStart);
  cmd.AddValue("monInterval", "monitor interval in s (1.0)",   monInterval);
  cmd.AddValue("quiet",       "do not show progress",          quiet);
  cmd.AddValue("queueTime",   "FIFO & queue time in ms (0)",   queueTime);
  cmd.AddValue("run",         "run number (0)",                run);
  cmd.AddValue("seed",        "random seed (0 = time)",        seed);
  cmd.AddValue("traceFile",   "trace file name",               traceFile);
  cmd.AddValue("stopTime",    "simulation duration in s (60)", stopTime);
  cmd.AddValue("setLD",       "use LD mode (0)",               setLD);
  cmd.AddValue("setABE",      "use ABE mode (0)",              setABE);
  cmd.AddValue("setPRIO",     "use PRIO mode (0)",             setPRIO);
  cmd.AddValue("scale",       "scale factor (1.0)",            scale);
  cmd.Parse(argc, argv);

  baseRate = baseRate * mega * scale;

  // fix seed and use run number
  if (!seed) seed = time(NULL);
  RngSeedManager::SetSeed(seed);
  SeedManager::SetSeed(seed);
  RngSeedManager::SetRun(run+1);
  SeedManager::SetRun(run+1);

  // LogComponentEnable("Socket", LOG_LEVEL_LOGIC);
  // LogComponentEnable("TcpSocket", LOG_LEVEL_LOGIC);
  // LogComponentEnable("TcpSocketBase", LOG_LEVEL_LOGIC);
  // LogComponentEnable("DelaySegmentFifo", LOG_LEVEL_LOGIC);
  // LogComponentEnable("Ipv4Header", LOG_LEVEL_LOGIC);
  // LogComponentEnable("Packet", LOG_LEVEL_LOGIC);
  // LogComponentEnable("PointToPointNetDevice", LOG_LEVEL_LOGIC);
  // LogComponentEnable("PointToPointChannel", LOG_LEVEL_LOGIC);
  // LogComponentEnable("Queue", LOG_LEVEL_LOGIC);

  Config::SetDefault ("ns3::TcpL4Protocol::SocketType", StringValue("ns3::TcpNewReno"));
	// Config::SetDefault ("ns3::TcpL4Protocol::SocketType", StringValue("ns3::TcpCubic"));
	// Config::SetDefault ("ns3::TcpSocket::RcvBufSize", UintegerValue(262144));
	// Config::SetDefault ("ns3::TcpSocket::SndBufSize", UintegerValue(262144));
	// Config::SetDefault ("ns3::TcpSocket::InitialSlowStartThreshold", UintegerValue(131072));
	// Config::SetDefault ("ns3::TcpSocket::InitialCwnd", UintegerValue(4));

  unsigned int nodeCount = parseConfigFile( configFile );

  // set up topology
  PointToPointHelper bottleneck;
  bottleneck.SetQueue("ns3::DropTailQueue");
  bottleneck.SetDeviceAttribute ("DataRate", DataRateValue(DataRate(baseRate)));
  bottleneck.SetChannelAttribute("Delay", TimeValue(MilliSeconds(linkDelay)));

  PointToPointHelper leaf;
  bottleneck.SetQueue("ns3::DropTailQueue");
  leaf.SetDeviceAttribute ("DataRate", DataRateValue(DataRate(baseRate * 1.1)));
  leaf.SetChannelAttribute("Delay", TimeValue(MilliSeconds(linkDelay)));

  topology = new PointToPointDumbbellHelper( nodeCount, leaf, nodeCount, leaf, bottleneck );

  NodeContainer edges, routers;
  routers.Add(topology->GetLeft());
  routers.Add(topology->GetRight());
  for (unsigned int i = 0; i < nodeCount; i += 1) {
    edges.Add(topology->GetLeft(i));
    edges.Add(topology->GetRight(i));
  }

  InternetStackHelper stack;
#if DSFDCE
  DceManagerHelper dceManager;
  LinuxStackHelper lstack;
  if (linuxStack) {
    GlobalValue::Bind("ChecksumEnabled", BooleanValue(true));
    dceManager.SetNetworkStack("ns3::LinuxSocketFdFactory", "Library", StringValue ("liblinux.so"));
    dceManager.Install(edges);
    lstack.Install(edges);

    lstack.SysctlSet(edges, ".net.ipv4.tcp_allowed_congestion_control", "reno cubic");
    lstack.SysctlSet(edges, ".net.ipv4.tcp_congestion_control", "cubic");
//    lstack.SysctlSet(edges, ".net.ipv4.tcp_moderate_rcvbuf", "0");
//    lstack.SysctlSet(edges, ".net.core.rmem_default", "375000");
//    lstack.SysctlSet(edges, ".net.core.wmem_default", "375000");
//    lstack.SysctlSet(edges, ".net.core.rmem_max", "375000");
//    lstack.SysctlSet(edges, ".net.core.wmem_max", "375000");
//    lstack.SysctlSet(edges, ".net.ipv4.tcp_rmem", "4096	87380	375000");
//    lstack.SysctlSet(edges, ".net.ipv4.tcp_wmem", "4096	16384	375000");
//    lstack.SysctlSet(edges, ".net.ipv4.tcp_rmem", "4096	87380 87380");
//    lstack.SysctlSet(edges, ".net.ipv4.tcp_wmem", "4096	16384 65536");
//    lstack.SysctlSet(routers, ".net.ipv4.conf.default.forwarding", "1");

//    LinuxStackHelper::SysctlGet (edges.Get(0), Seconds (1.0), ".net.ipv4.tcp_moderate_rcvbuf", &Sysctl);
//    LinuxStackHelper::SysctlGet (edges.Get(0), Seconds (1.0), ".net.core.rmem_default", &Sysctl);
//    LinuxStackHelper::SysctlGet (edges.Get(0), Seconds (1.0), ".net.core.wmem_default", &Sysctl);
//    LinuxStackHelper::SysctlGet (edges.Get(0), Seconds (1.0), ".net.core.rmem_max", &Sysctl);
//    LinuxStackHelper::SysctlGet (edges.Get(0), Seconds (1.0), ".net.core.wmem_max", &Sysctl);
//    LinuxStackHelper::SysctlGet (edges.Get(0), Seconds (1.0), ".net.ipv4.tcp_rmem", &Sysctl);
//    LinuxStackHelper::SysctlGet (edges.Get(0), Seconds (1.0), ".net.ipv4.tcp_wmem", &Sysctl);
//    LinuxStackHelper::SysctlGet (edges.Get(0), Seconds (1.0), ".net.ipv4.tcp_window_scaling", &Sysctl);
//    LinuxStackHelper::SysctlGet (edges.Get(0), Seconds (1.0), ".net.ipv4.tcp_allowed_congestion_control", &Sysctl);
//    LinuxStackHelper::SysctlGet (edges.Get(0), Seconds (1.0), ".net.ipv4.tcp_available_congestion_control", &Sysctl);
//    LinuxStackHelper::SysctlGet (edges.Get(0), Seconds (1.0), ".net.ipv4.tcp_congestion_control", &Sysctl);

  } else
#endif
    stack.Install(edges);

  stack.Install(routers);

  topology->AssignIpv4Addresses( Ipv4AddressHelper ("10.0.1.0", "255.255.255.0"),
                                 Ipv4AddressHelper ("11.0.1.0", "255.255.255.0"),
                                 Ipv4AddressHelper ("12.0.1.0", "255.255.255.0") );

  // setup static routing tables
  Ipv4GlobalRoutingHelper::PopulateRoutingTables();

#if DSFDCE
  if (linuxStack) {
    // setup static routing tables
    LinuxStackHelper::PopulateRoutingTables();
  }
  DceApplicationHelper process;
#endif

  list<DSF_App*> appList;
  unsigned int nodeNum = 0;
  for (list<DSF_Desc>::iterator it = descList.begin(); it != descList.end(); ++it) {
    for (unsigned int i = 0; i < it->count; i += 1) {
      DSF_App* a = new DSF_App(*it, nodeNum);
      appList.push_back(a);
      nodeNum += 1;
    }
  }
  if (nodeNum != nodeCount) terminate("PANIC");

  // get device in forwarding path of bottleneck
  Ptr<Node> n = topology->GetLeft();
  Ptr<PointToPointNetDevice> device = n->GetDevice(0)->GetObject<PointToPointNetDevice>();

  // set up DSF queue
  Ptr<DSFQueue> dsf = new DSFQueue();
  dsf->setLinkRate(baseRate);
  // switch to FIFO mode
  if (queueTime) dsf->setQueueLimit((double(queueTime) / 8000) * baseRate);
  // set up classes
  for ( list<DSFClass>::iterator it = classList.begin(); it != classList.end(); ++ it ) {
    dsf->AddClass(*it);
  }
  // switch to ABE or PRIO mode
  if (setLD) dsf->setLD();
  if (setABE) dsf->setABE();
  if (setPRIO) dsf->setPRIO();
  // trace and monitor files
  if (!compFile.empty())  completionFile.open(compFile.c_str());
  if (!monFile.empty())   dsf->startMonitor(monFile, monFile + ".pkt", monFile + ".drop", monStart, monInterval, classList.size());
  if (!traceFile.empty()) dsf->setTraceFile(traceFile);
  // install DSF queue
  device->SetQueue(dsf);

  FlowMonitorHelper flowMonHelper;
  Ptr<FlowMonitor> flowMon;
  if (!flowmonFile.empty()) flowMon = flowMonHelper.InstallAll();

  // start simulation
  if (!quiet) Simulator::Schedule(Seconds(stopTime) / 100, &printTime);
  Simulator::Stop(Seconds(stopTime) + Seconds(0.000001));
  Simulator::Run();
  if (!quiet) { printTime(); cerr << endl; }

  if (!monFile.empty())   dsf->stopMonitor();
  if (!flowmonFile.empty()) flowMon->SerializeToXmlFile(flowmonFile, false, false);

  Simulator::Destroy();

  for (list<DSF_Desc>::iterator it = descList.begin(); it != descList.end(); ++it) {
    if (it->trace) delete it->trace;
  }

  for (list<DSF_App*>::iterator it = appList.begin(); it != appList.end(); ++it) {
    delete *it;
  }

  return 0;
}
