#include "ns3/core-module.h"
#include "ns3/dsf-queue.h" 
#include "ns3/ipv4-header.h"
#include "ns3/ppp-header.h"
#include "ns3/log.h"

#include <map>

namespace ns3 {

#define DSFPLEN(p) p->GetSize() 
#define DSFNOW (Now().GetDouble() / 1000000000)
#define DSFASSERT(cond) NS_ASSERT(cond)
#define DSFDROP(caller, p) DSFQueue::drop(caller, p)
#define DSFLDROP(caller, p) DSFQueue::drop(caller, p, true)
#define DSFTRACE(caller, b) DSFQueue::trace(caller, b)
#define DSFRESET(caller) DSFQueue::reset(caller)
#include "ns3/dsfbase2.h"

NS_LOG_COMPONENT_DEFINE("DelaySegmentFIFO");

TypeId DSFQueue::GetTypeId(void) {
  static TypeId tid = TypeId ("ns3::DSFQueue")
  .SetParent<Queue> ()
  .AddConstructor<DSFQueue> ();
  return tid;
}

NS_OBJECT_ENSURE_REGISTERED(DSFQueue);

DSFQueue::DSFQueue() : monCount(0), linkRate(0), initialized(false) {
  NS_LOG_FUNCTION_NOARGS();
  reset(this);
}

DSFQueue::~DSFQueue() {
  NS_LOG_FUNCTION_NOARGS();
}

void DSFQueue::setLD() {
  base.setLD();
}

void DSFQueue::setABE() {
  base.setABE();
}

void DSFQueue::setPRIO() {
  base.setPRIO();
}

void DSFQueue::setTraceFile( const string& filename ) {
  traceFile.open( filename.c_str() );
  if (traceFile.good()) base.traceflag = true;
  else base.traceflag = false;
}

void DSFQueue::startMonitor( const string& filename, const string& filenamePkt, const string& filenameDrop, double start, double interval, unsigned int nodeCount ) {
  monFile.open( filename.c_str() );
  monFilePkt.open( filenamePkt.c_str() );
  monFileDrop.open( filenameDrop.c_str() );
  monInterval = interval;
  NS_ASSERT(nodeCount && nodeCount <= monMax);
  monCount = nodeCount;
  Simulator::Schedule(Seconds(start), &DSFQueue::monitor, this);
}

inline bool DSFQueue::classify(Ptr<Packet> p, uint32_t& sc, uint32_t& tc) {
  bool success = false;
  PppHeader ppp;
  Ipv4Header hdr;
  if (p->RemoveHeader(ppp)) {
    if (p->PeekHeader(hdr)) {
      success = true;
      tc = sc = (hdr.GetTos() >> 2);
    }
    p->AddHeader(ppp);
  }
  if (monCount) NS_ASSERT(tc < monCount);
  return success;
}

void DSFQueue::drop(DSFQueue* This, Ptr<Packet> p, bool late) {
  if (This->monCount) {
    uint32_t sc, tc;
    if (This->classify(p, sc, tc)) {
      if (This->svctrain) This->monFilePkt << This->svctrain << endl;
      This->svctrain = 0;
      This->droptrain += 1;
      This->mon[tc].drop(p->GetSize(), late);
    }
  }
  This->Drop(p);
}

void DSFQueue::trace( DSFQueue* This, char* buf ) {
  This->traceFile << buf << endl;
}

bool DSFQueue::DoEnqueue(Ptr<Packet> p) {
  NS_LOG_FUNCTION (this << p);
  if (!initialized) {
    base.reset();
    initialized = true;
  }
  uint32_t sc, tc;
  if (!classify(p, sc, tc)) {
    cerr << "Unmarked packet dropped" << endl;
    return false;
  }
  if (monCount) mon[tc].arrival(p->GetSize());
  return base.enqueue(p, sc, this);
}

Ptr<Packet> DSFQueue::DoDequeue(void) {
  double delay = 0;
  Ptr<Packet> p = base.dequeue( delay, this );
  if (p && monCount) {
    uint32_t sc, tc;
    if (classify(p, sc, tc)) {
      if (droptrain) monFileDrop << droptrain << endl;
      droptrain = 0;
      svctrain += 1;
      mon[tc].service(p->GetSize(), delay);
    }
  }
  return p;
}

Ptr<const Packet> DSFQueue::DoPeek(void) const {
  terminate("DSFQueue::DoPeek not implemented");
  return NULL;
}

void DSFQueue::stopMonitor() {
  if (!monCount) return;
  for (unsigned int i = 0; i < monCount; i += 1) {
    char buf[512];
    char* b = buf + sprintf(buf, "T  SUMMARY SC %-3i", i);
    mon[i].summary();
    tprint(b, mon[i].tot);
    monFile << buf << endl;
  }
  monFile.close();
  monFilePkt.close();
  monFileDrop.close();
}

inline void DSFQueue::tprint(char* buf, Monitor::Data& d) {
  uint64_t tot = d.pservice + d.pdropped;
  sprintf( buf, " %10lu %10lu %8.6f %8.6f %8.6f %8.6f",
    d.barrival, d.bservice, d.delayMax,
    d.pservice ? (d.delaySum / d.pservice) : 0,
    tot ? (double(d.pdropped) / tot) : 0,
    tot ? (double(d.ldropped) / tot) : 0);
}

inline  void DSFQueue::monitor() {
  if (!monCount) return;
  for (unsigned int i = 0; i < monCount; i += 1) {
    char buf[512];
    char* b = buf + sprintf(buf, "t %8.3f sc %-3i", Simulator::Now().GetSeconds(), i);
    tprint(b, mon[i].cur);
    monFile << buf << endl;
    mon[i].interval();
  }
  Simulator::Schedule(Seconds(monInterval), &DSFQueue::monitor, this);
}

}; // namespace
