## configuration
load = 1.00
precision = 1024        # floating point precision [bits]
delayTargets = 2^(0:13) # D1: delay target for class 1
Qlen = 10000            # D2: delay target for class 2 (total queue length) [pakets]
rates = c(.0001,.001,.01,.1,.3,.5,.7,.9,.99,.999,.9999) # relative rate of class 1

## process command-line arguments (file name)
args <- commandArgs(TRUE)
outputfileChain = args[1]
outputfileLoss = args[2]

## check parts of configuration ("assert")
stopifnot(max(delayTargets) < Qlen)
stopifnot(load > 0 & load < 1.6) # numerically unstable for load > 1.6 (tested until 1.8)

## R-library imports
require(foreach)
require(data.table)
require(ggplot2)
require(Rmpfr)
require(doParallel)
registerDoParallel()

## precision helper
mp = function(x)
    mpfr(x,precBits = precision)

# vertical chain steady-state probability
DSFvert = function(a,d,D2,j) {
     #0-state's
    if(a==2*d)
        return(1/(D2+1))
    piQ0 = ( 1 - a/(2*d) ) /
        ( 1 - (a/(2*d))^(D2+1) )
    return( (a/(2*d))^j * piQ0 )
}

# horizontal chain steady-state probability
DSFhor = function(a,d,D1,i) {
    #0-state's 
    if(a==d)
        return(1/(D1+1))
    pi0 = ( 1 - a/d ) /
        ( 1 - (a/d)^(D1+1) )
    return( (a/d)^i * pi0 )
}

## loss probability for DSF
## class 1
DSFlossCl1 = function(p1,p2,D1,D2) {
    stopifnot(p1 > 0 & p1 <= 1 & p2 > 0 & p2 <= 1)
    # event additional packet in queue
    a = p1*p2
    # event no arrival
    d = (1 - p1) * (1 - p2)
    # loss formula: for all verticals take last horizontal
#    return ( a/2 * sum(DSFvert(a,d,D2,seq(mp(0),D2,mp(1)))) * DSFhor(a,d,D1,D1) )
    # sum always adds to 1
    return ( a * 1 * DSFhor(a,d,D1,D1) )
}

## class 2
DSFlossCl2 = function(p1,p2,D1,D2) {
    stopifnot(p1 > 0 & p1 <= 1 & p2 > 0 & p2 <= 1)
    # event additional packet in queue
    a = p1*p2
    # event no arrival
    d = (1 - p1) * (1 - p2)
    # loss formula: for all verticals take last horizontal
    return ( a * DSFvert(a,d,D2,D2) * DSFhor(a,d,D1,D1) )
}

## compute loop
results=
    foreach(rate = rates, .combine=rbind, .inorder=FALSE, .errorhandling="stop") %:%
    foreach(dTarget = delayTargets, .combine=rbind, .inorder=FALSE, .errorhandling="stop") %dopar% {
            data.table('rate'=rate, 'dTarget'=dTarget,
                       'tp1'=as.numeric(1-DSFlossCl1(mp(rate*load), mp((1-rate)*load), mp(dTarget), mp(Qlen))),
                       'tp2'=as.numeric(1-DSFlossCl2(mp(rate*load), mp((1-rate)*load), mp(dTarget), mp(Qlen)))
                       )
        }

## store results
results[,tp1:=format(tp1,scientific = FALSE)]
results[,tp2:=format(tp2,scientific = FALSE)]
write.table(results, outputfileChain, quote=F, col.names=F, row.names=F)




## compute loop 2
load = 0.9999
rates = c(0.001,seq(0.05,0.95,0.01),0.999)
dTarget = 1

## compute loop for fixed delay target of 2 packets
results2=
    foreach(rate = rates, .combine=rbind, .inorder=FALSE, .errorhandling="stop") %dopar% {
        data.table('rate'=rate, 'dTarget'=dTarget,
                   'l1'=as.numeric(DSFlossCl1(mp(rate*load), mp((1-rate)*load), mp(dTarget), mp(Qlen))),
                   'l2'=as.numeric(DSFlossCl2(mp(rate*load), mp((1-rate)*load), mp(dTarget), mp(Qlen)))
                       )
        }

results2[,tl:=( ( l1*rate*load + l2*(1-rate)*load ) )]

## store results
write.table(results2, outputfileLoss, quote=F, col.names=F, row.names=F)
