## configuration
load = 0.9999
precision = 1024        # floating point precision [bits]
delayTargets = c(2^(0:13),3000,5000,6000,7000,9000,9900) # D1: delay target for class 1
Qlen = 10000            # D2: delay target for class 2 (total queue length) [pakets]
rates = c(0.5)

## process command-line arguments (file name)
args <- commandArgs(TRUE)
outputfileLate = args[1]
outputfileLoss = args[2]

## check parts of configuration ("assert")
stopifnot(max(delayTargets) < Qlen)
stopifnot(load > 0 & load < 1.6) # numerically unstable for load > 1.6 (tested until 1.8)

## R-library imports
require(foreach)
require(data.table)
require(ggplot2)
require(Rmpfr)
require(doParallel)
registerDoParallel()

## precision helper
mp = function(x)
    mpfr(x,precBits = precision)

## steady-state probability for FIFO Markov chain
FIFOprob = function(a,d,D2,i) {
    stopifnot(!(a==d))
    pi0 = ( a/d - 1 ) /
        ( (a/d)^(D2+1) - 1)
    return( (a/d)^i * pi0 )
}

## late packet probability for FIFO
FIFOlate = function(p1,p2,D1,D2) {
    stopifnot(p1 > 0 & p1 <= 1 & p2 > 0 & p2 <= 1)
    # event additional packet in queue
    a = p1*p2
    # event no arrival
    d = (1 - p1) * (1 - p2)
    if(a==d)
        return((D2-(D1+1))*1/(D2+1))
    return( sum(FIFOprob(a, d, D2, seq(D1+mp(1),D2,mp(1)))) )
}

## total loss probability for FIFO
FIFOloss = function(p1,p2,D2) {
    stopifnot(p1 > 0 & p1 <= 1 & p2 > 0 & p2 <= 1)
    # event additional packet in queue
    a = p1*p2
    # event no arrival
    d = (1 - p1) * (1 - p2)
    if(a==d)
        return(a * 1/(D2+1))
    return( a* FIFOprob(a, d, D2, D2) )
}

## compute loop
results=
    foreach(rate = rates, .combine=rbind, .inorder=FALSE, .errorhandling="stop") %:%
    foreach(dTarget = delayTargets, .combine=rbind, .inorder=FALSE, .errorhandling="stop") %dopar% {
        data.table('rate'=rate, 'dTarget'=dTarget,
                   'late'=as.numeric(FIFOlate(mp(rate*load), mp((1-rate)*load), mp(dTarget), mp(Qlen))),
                   'loss'=as.numeric(FIFOloss(mp(rate*load), mp((1-rate)*load), mp(dTarget)))
                   )
        }


## store results
results[,late:=format(late,scientific = FALSE)]
write.table(results[,list(rate,dTarget,late)], outputfileLate, quote=F, col.names=F, row.names=F)
results[,loss:=format(loss,scientific = FALSE)]
write.table(results[,list(rate,Qlen=dTarget,loss)], outputfileLoss, quote=F, col.names=F, row.names=F)


