## configuration
load = 0.999999999999999999
precision = 2048        # floating point precision [bits]
delayTargets = 2^(0:13) # D1: delay target for class 1
Qlen = 10000            # D2: delay target for class 2 (total queue length) [pakets]
rates = c(.0001,.001,.01,0.1,0.2,.3,0.4,.5,0.6,0.7,0.8,.9,.99,.999,.9999) # relative rate of class 1

## process command-line arguments (file name)
args <- commandArgs(TRUE)
outputfolder = args[1]

## check parts of configuration ("assert")
stopifnot(max(delayTargets) < Qlen)
stopifnot(load > 0 & load < 1.6) # numerically unstable for load > 1.6 (tested until 1.8)

## R-library imports
require(foreach)
require(data.table)
require(ggplot2)
require(Rmpfr)
require(doParallel)
registerDoParallel()

## precision helper
mp = function(x)
    mpfr(x,precBits = precision)




# vertical chain steady-state probability
DSFvert = function(a,d,D2,j) {
     #0-state's
    if(a==2*d)
        return(1/(D2+1))
    piQ0 = ( 1 - a/(2*d) ) /
        ( 1 - (a/(2*d))^(D2+1) )
    return( (a/(2*d))^j * piQ0 )
}

# horizontal chain steady-state probability
DSFhor = function(a,d,D1,i) {
    #0-state's 
    if(a==d)
        return(1/(D1+1))
    pi0 = ( 1 - a/d ) /
        ( 1 - (a/d)^(D1+1) )
    return( (a/d)^i * pi0 )
}

## loss probability for DSF
## class 1
DSFlossCl1 = function(p1,p2,D1,D2) {
    stopifnot(p1 > 0 & p1 <= 1 & p2 > 0 & p2 <= 1)
    # event additional packet in queue
    a = p1*p2
    # event no arrival
    d = (1 - p1) * (1 - p2)
    # loss formula: for all verticals take last horizontal
#    return ( a/2 * sum(DSFvert(a,d,D2,seq(mp(0),D2,mp(1)))) * DSFhor(a,d,D1,D1) )
    # sum always adds to 1
    return ( a * 1 * DSFhor(a,d,D1,D1) )
}

## class 2
DSFlossCl2 = function(p1,p2,D1,D2) {
    stopifnot(p1 > 0 & p1 <= 1 & p2 > 0 & p2 <= 1)
    # event additional packet in queue
    a = p1*p2
    # event no arrival
    d = (1 - p1) * (1 - p2)
    # loss formula: for all verticals take last horizontal
    return ( a * DSFvert(a,d,D2,D2) * DSFhor(a,d,D1,D1) )
}

## compute loop for DSF TII
results=
    foreach(rate = rates, .combine=rbind, .inorder=FALSE, .errorhandling="stop") %:%
    foreach(dTarget = delayTargets, .combine=rbind, .inorder=FALSE, .errorhandling="stop") %dopar% {
            data.table('rate'=rate, 'dTarget'=dTarget,
                       'l1'=rate*load,
                       'tp1'=as.numeric(mp(rate*load)-0.5*DSFlossCl1(mp(rate*load), mp((1-rate)*load), mp(dTarget), mp(Qlen))),
                       'l2'=(1-rate)*load,
                       'tp2'=as.numeric(mp((1-rate)*load)-0.5*DSFlossCl2(mp(rate*load), mp((1-rate)*load), mp(dTarget), mp(Qlen)))
                       )
        }

results[,r1:=tp1/l1]
results[,r2:=tp2/l2]
results[,tii:=1 - (  (r1+r2)^2 / (2* ( (r1)^2+(r2)^2 ) )  )]

outputfile = paste(outputfolder,"dsfMarkov.table",sep="/")
## store results.dsf
results[,tii:=format(tii,scientific = FALSE)]
write.table(results[,list(rate,dTarget,tii)], outputfile, quote=F, col.names=F, row.names=F)










## TII for VIFQ
## class 1
VIFQlossCl1 = function(r1,r2,N1,N) {
    stopifnot(r1 > 0 & r1 <= 1 & r2 > 0 & r2 <= 1)
    # notation from tech report
    q <- (r1*r2) / ( (1-r1)*(1-r2) )
    s <- (1-VIFQlossCl2(r1,r2,N1,N))*r1
    if(q == 1) {
        return(
            1 / (
            (r1^2*( (1-s)/(1-r1) )^N1)/(s*(r1-s)) - r1/(r1-s) + r1^2*(1-s)^N1
            ) * (N - N1) / (N+1)
            )
    } else {
        return(
            1 / (
            (r1^2*( (1-s)/(1-r1) )^N1)/(s*(r1-s)) - r1/(r1-s) + r1^2*(1-s)^N1
            ) * q^(N1+1) * (1-q^(N-N1))/(1-q^(N+1))
            )
    }
}

## class 2
VIFQlossCl2 = function(r1,r2,N1,N) {
    stopifnot(r1 > 0 & r1 <= 1 & r2 > 0 & r2 <= 1)
    # simple fifo queue loss probability
    q <- (r1*r2) / ( (1-r1)*(1-r2) )
    if (q == 1) {
        return(1/(N+1))
    }
    pi0 <- (1-q)/(1-q^(N+1))
    return(q^N*pi0)
}

results.vifq=
    foreach(rate = rates, .combine=rbind, .inorder=FALSE, .errorhandling="stop") %:%
    foreach(dTarget = delayTargets, .combine=rbind, .inorder=FALSE, .errorhandling="stop") %dopar% {
            data.table('rate'=rate, 'dTarget'=dTarget,
                       'l1'=rate*load,
                       'tp1'=as.numeric(mp(rate*load)-
                                            mp(rate*load)*VIFQlossCl1(mp(rate*load), mp((1-rate)*load), mp(dTarget), mp(Qlen))-
                           mp(rate*load)*VIFQlossCl2(mp(rate*load), mp((1-rate)*load), mp(dTarget), mp(Qlen))),
                       'l2'=(1-rate)*load,
                       'tp2'=as.numeric(mp((1-rate)*load)-mp((1-rate)*load)*VIFQlossCl2(mp(rate*load), mp((1-rate)*load), mp(dTarget), mp(Qlen)))
                       )
        }

results.vifq[,r1:=tp1/l1]
results.vifq[,r2:=tp2/l2]
results.vifq[,tii:=1 - (  (r1+r2)^2 / (2* ( (r1)^2+(r2)^2 ) )  )]


outputfile = paste(outputfolder,"vifqMarkov.table",sep="/")

## store results.vifq
results.vifq[,tii:=format(tii,scientific = FALSE)]
write.table(results.vifq[,list(rate,dTarget,tii)], outputfile, quote=F, col.names=F, row.names=F)


